@@ -5,7 +5,7 @@ import random
 import numpy as np
 
 from sureal.tools.misc import empty_object, get_unique_sorted_list
-from sureal.tools.decorator import memoized as persist
+from sureal.tools.decorator import memoized as persist, override
 from sureal.tools.misc import get_unique_sorted_list
 
 __copyright__ = "Copyright 2016-2018, Netflix, Inc."
@@ -30,14 +30,18 @@ class DatasetReader(object):
                     format(self.num_ref_videos, cid)
 
         # assert dis_video content_id is content_ids
-        for dis_video in self.dataset.dis_videos:
+        for dis_video in self.dis_videos:
             assert dis_video['content_id'] in cids, \
                 "dis_video of content_id {content_id}, asset_id {asset_id} must have content_id in {cids}".format(
                     content_id=dis_video['content_id'], asset_id=dis_video['asset_id'], cids=cids)
 
+    @property
+    def dis_videos(self):
+        return self.dataset.dis_videos
+
     @property
     def num_dis_videos(self):
-        return len(self.dataset.dis_videos)
+        return len(self.dis_videos)
 
     @property
     def num_ref_videos(self):
@@ -53,11 +57,11 @@ class DatasetReader(object):
 
     @property
     def asset_ids(self):
-        return list(set(map(lambda dis_video: dis_video['asset_id'], self.dataset.dis_videos)))
+        return list(set(map(lambda dis_video: dis_video['asset_id'], self.dis_videos)))
 
     @property
     def content_id_of_dis_videos(self):
-        return list(map(lambda dis_video: dis_video['content_id'], self.dataset.dis_videos))
+        return list(map(lambda dis_video: dis_video['content_id'], self.dis_videos))
 
     @property
     def _contentid_to_refvideo_map(self):
@@ -71,7 +75,7 @@ class DatasetReader(object):
         d = self._contentid_to_refvideo_map
         return list(map(
             lambda dis_video: d[dis_video['content_id']]['path'] == dis_video['path'],
-            self.dataset.dis_videos
+            self.dis_videos
         ))
 
     @property
@@ -107,7 +111,7 @@ class DatasetReader(object):
 class RawDatasetReader(DatasetReader):
     """
     Reader for a subjective quality test dataset with raw scores (dis_video must
-    has key of 'os' (opinion score)).
+    have key of 'os' (opinion score)).
     """
 
     def _assert_dataset(self):
@@ -119,63 +123,109 @@ class RawDatasetReader(DatasetReader):
 
         # assert each dis_video dict has key 'os' (opinion score), and must
         # be iterable (list, tuple or dictionary)
-        for dis_video in self.dataset.dis_videos:
+        for dis_video in self.dis_videos:
             assert 'os' in dis_video, "dis_video must have key 'os' (opinion score)"
             assert isinstance(dis_video['os'], (list, tuple, dict))
 
         # make sure each dis video has equal number of observers
         if (
-                    isinstance(self.dataset.dis_videos[0]['os'], list) or
-                    isinstance(self.dataset.dis_videos[0]['os'], tuple)
+                    isinstance(self.dis_videos[0]['os'], list) or
+                    isinstance(self.dis_videos[0]['os'], tuple)
         ):
-            num_observers = len(self.dataset.dis_videos[0]['os'])
-            for dis_video in self.dataset.dis_videos[1:]:
+            num_observers = len(self.dis_videos[0]['os'])
+            for dis_video in self.dis_videos[1:]:
                 assert num_observers == len(dis_video['os']), \
                     "expect number of observers {expected} but got {actual} for {dis_video}".format(
                         expected=num_observers, actual=len(dis_video['os']), dis_video=str(dis_video))
 
-    @property
-    def num_observers(self):
+    def _get_num_observers(self):
         if (
-                    isinstance(self.dataset.dis_videos[0]['os'], list) or
-                    isinstance(self.dataset.dis_videos[0]['os'], tuple)
+                    isinstance(self.dis_videos[0]['os'], list) or
+                    isinstance(self.dis_videos[0]['os'], tuple)
         ):
-            return len(self.dataset.dis_videos[0]['os'])
-        elif isinstance(self.dataset.dis_videos[0]['os'], dict):
+            return len(self.dis_videos[0]['os'])
+        elif isinstance(self.dis_videos[0]['os'], dict):
             list_observers = self._get_list_observers()
             return len(list_observers)
         else:
             assert False, ''
 
+    @property
+    def num_observers(self):
+        return self._get_num_observers()
+
     def _get_list_observers(self):
 
-        for dis_video in self.dataset.dis_videos:
+        for dis_video in self.dis_videos:
             assert isinstance(dis_video['os'], dict)
 
         list_observers = []
-        for dis_video in self.dataset.dis_videos:
+        for dis_video in self.dis_videos:
             list_observers += dis_video['os'].keys()
 
         return get_unique_sorted_list(list_observers)
 
-    @property
-    def opinion_score_2darray(self):
+    def _get_max_repetitions(self):
+        """get the maximum number of times any observer evaluated any stimulus. For example, if all of the observers
+        evaluated each stimulus only once, the function returns 1.
         """
-        2darray storing raw opinion scores, with first dimension the number of
-        distorted videos, second dimension the number of observers
-        """
-        score_mtx = float('NaN') * np.ones([self.num_dis_videos, self.num_observers])
+        max_reps = 1
+        if (
+                    isinstance(self.dis_videos[0]['os'], list) or
+                    isinstance(self.dis_videos[0]['os'], tuple)
+        ):
+            for dis_video in self.dis_videos:
+                for idx_obs in range(self.num_observers):
+                    scores = dis_video['os'][idx_obs]
+                    if isinstance(scores, list) or isinstance(scores, tuple):
+                        if len(scores) > max_reps:
+                            max_reps = len(scores)
 
-        if isinstance(self.dataset.dis_videos[0]['os'], list) \
-                or isinstance(self.dataset.dis_videos[0]['os'], tuple):
-            for i_dis_video, dis_video in enumerate(self.dataset.dis_videos):
-                score_mtx[i_dis_video, :] = dis_video['os']
-        elif isinstance(self.dataset.dis_videos[0]['os'], dict):
+        elif isinstance(self.dis_videos[0]['os'], dict):
+            for dis_video in self.dis_videos:
+                list_observers = dis_video['os'].keys()
+
+                for observer in list_observers:
+                    scores = dis_video['os'][observer]
+                    if isinstance(scores, list) or isinstance(scores, tuple):
+                        if len(scores) > max_reps:
+                            max_reps = len(scores)
+        else:
+            assert False, ''
+
+        return max_reps
+
+    @property
+    def max_repetitions(self):
+        return self._get_max_repetitions()
+
+    @property
+    def opinion_score_3darray(self):
+        """
+        3darray storing raw opinion scores, with first dimension the number of
+        distorted videos, second dimension the number of observers, and third the maximum number of repetitions
+        """
+        score_mtx = float('NaN') * np.ones([self.num_dis_videos, self._get_num_observers(), self._get_max_repetitions()])
+
+        if isinstance(self.dis_videos[0]['os'], list) \
+                or isinstance(self.dis_videos[0]['os'], tuple):
+            for i_dis_video, dis_video in enumerate(self.dis_videos):
+                for i_observer in range(self.num_observers):
+                    if isinstance(dis_video['os'][i_observer], list) or isinstance(dis_video['os'][i_observer], tuple):
+                        reps = len(dis_video['os'][i_observer])
+                    else:
+                        reps = 1
+                    score_mtx[i_dis_video, i_observer, :reps] = dis_video['os'][i_observer]
+        elif isinstance(self.dis_videos[0]['os'], dict):
             list_observers = self._get_list_observers()
-            for i_dis_video, dis_video in enumerate(self.dataset.dis_videos):
+            for i_dis_video, dis_video in enumerate(self.dis_videos):
                 for i_observer, observer in enumerate(list_observers):
                     if observer in dis_video['os']:
-                        score_mtx[i_dis_video, i_observer] = dis_video['os'][observer]
+                        if isinstance(dis_video['os'][observer], list) or isinstance(dis_video['os'][observer], tuple):
+                            reps = len(dis_video['os'][observer])
+                        else:
+                            reps = 1
+                        score_mtx[i_dis_video, i_observer, :reps] = dis_video['os'][observer]
         else:
             assert False
         return score_mtx
@@ -189,12 +239,18 @@ class RawDatasetReader(DatasetReader):
 
         # dis_videos: use input aggregate scores
         dis_videos = []
-        assert len(self.dataset.dis_videos) == len(aggregate_scores)
-        for dis_video, score in zip(self.dataset.dis_videos, aggregate_scores):
+        assert len(self.dis_videos) == len(aggregate_scores)
+        for dis_video, score in zip(self.dis_videos, aggregate_scores):
             dis_video2 = copy.deepcopy(dis_video)
             if 'os' in dis_video2: # remove 'os' - opinion score
                 del dis_video2['os']
-            dis_video2['groundtruth'] = score
+            if isinstance(score, np.ndarray):
+                if len(score) == 1:
+                    dis_video2['groundtruth'] = float(score)
+                else:
+                    dis_video2['groundtruth'] = list(score)
+            else:
+                dis_video2['groundtruth'] = score
             dis_videos.append(dis_video2)
 
         # add scores std if available
@@ -248,7 +304,7 @@ class RawDatasetReader(DatasetReader):
 
         # dis_videos: use input aggregate scores
         dis_videos = []
-        for dis_video, quality_score in zip(self.dataset.dis_videos, quality_scores):
+        for dis_video, quality_score in zip(self.dis_videos, quality_scores):
             assert 'os' in dis_video
 
             # quality_score should be a 1-D array with (processed) per-subject scores
@@ -265,7 +321,13 @@ class RawDatasetReader(DatasetReader):
                 dis_video2 = copy.deepcopy(dis_video)
                 if 'os' in dis_video2: # remove 'os' - opinion score
                     del dis_video2['os']
-                dis_video2['groundtruth'] = persubject_score
+                if isinstance(persubject_score, np.ndarray):
+                    if len(persubject_score) == 1:
+                        dis_video2['groundtruth'] = float(persubject_score)
+                    else:
+                        dis_video2['groundtruth'] = list(persubject_score)
+                else:
+                    dis_video2['groundtruth'] = persubject_score
                 dis_videos.append(dis_video2)
         newone.dis_videos = dis_videos
 
@@ -305,7 +367,7 @@ class RawDatasetReader(DatasetReader):
         if sampling_rate is not None:
             assert np.isscalar(sampling_rate) and 0.0 <= sampling_rate
         if per_asset_sampling_rates is not None:
-            assert len(per_asset_sampling_rates) == len(self.dataset.dis_videos)
+            assert len(per_asset_sampling_rates) == len(self.dis_videos)
             for per_asset_sampling_rate in per_asset_sampling_rates:
                 assert np.isscalar(per_asset_sampling_rate) and 0.0 <= per_asset_sampling_rate
 
@@ -313,7 +375,7 @@ class RawDatasetReader(DatasetReader):
         if cointoss_rate is not None:
             assert np.isscalar(cointoss_rate) and 0.0 <= cointoss_rate <= 1.0
         if per_asset_cointoss_rates is not None:
-            assert len(per_asset_cointoss_rates) == len(self.dataset.dis_videos)
+            assert len(per_asset_cointoss_rates) == len(self.dis_videos)
             for cointoss_rate_ in per_asset_cointoss_rates:
                 assert np.isscalar(cointoss_rate_) and 0.0 <= cointoss_rate_ <= 1.0
 
@@ -321,16 +383,16 @@ class RawDatasetReader(DatasetReader):
         if noise_level is not None:
             assert np.isscalar(noise_level) and 0.0 <= noise_level
         if per_asset_noise_levels is not None:
-            assert len(per_asset_noise_levels) == len(self.dataset.dis_videos)
+            assert len(per_asset_noise_levels) == len(self.dis_videos)
             for noise_level_ in per_asset_noise_levels:
                 assert np.isscalar(noise_level_) and 0.0 <= noise_level_
 
         if per_asset_mean_scores is not None:
-            assert len(per_asset_mean_scores) == len(self.dataset.dis_videos)
+            assert len(per_asset_mean_scores) == len(self.dis_videos)
             for mean_score_ in per_asset_mean_scores:
                 assert np.isscalar(mean_score_)
 
-        dis_videos = self.dataset.dis_videos
+        dis_videos = self.dis_videos
         if isinstance(dis_videos[0]['os'], dict):
             pass
         elif isinstance(dis_videos[0]['os'], (list, tuple)):
@@ -459,15 +521,102 @@ class RawDatasetReader(DatasetReader):
         pc_dataset = self.to_pc_dataset(**kwargs)
         self.write_out_dataset(pc_dataset, dataset_filepath)
 
+    def to_combined_overlap_dataset(self, second_dataset_reader, **kwargs):
+        """
+        A function to find an overlap between self.dataset and the second dataset and combine the scores for the
+        overlapping videos. The overlap is determined by matching paths.
+        """
+        assert isinstance(second_dataset_reader, RawDatasetReader), 'RawDatasetReader can only be combined with ' \
+                                                                    'another RawDatasetReader'
+
+        # make both of the datasets dictionary style
+        first_dataset = self.to_dictionary_style_dataset()
+        second_dataset = second_dataset_reader.to_dictionary_style_dataset()
+
+        newone = self._prepare_new_dataset(kwargs)
+        newone.ref_videos = []
+        newone.dis_videos = []
+
+        content_ids_in = []
+        for vid1 in first_dataset.dis_videos:
+            for vid2 in second_dataset.dis_videos:
+                if vid1['path'] == vid2['path']:  # TODO: Perhaps worth modifying so that the path does not need to match exactly
+
+                    # if the reference video isn't already in the newone.ref_videos then add it
+                    if vid1['content_id'] not in content_ids_in:
+                        content_ids_in.append(vid1['content_id'])
+                        new_content_id = content_ids_in.index(vid1['content_id'])
+                        for ref in first_dataset.ref_videos:
+                            if ref['content_id'] == vid1['content_id']:
+                                new_ref = copy.deepcopy(ref)
+                                new_ref['content_id'] = new_content_id
+                                newone.ref_videos.append(new_ref)
+                                break
+                    else:
+                        new_content_id = content_ids_in.index(vid1['content_id'])
+
+                    new_dis_video = copy.deepcopy(vid1)
+                    new_dis_video['content_id'] = new_content_id
+
+                    new_dis_video['os'].update(vid2['os'])  # combined the 'os's but overrode when subject is in both
+
+                    for key in vid2['os'].keys():
+                        if key in vid1['os'].keys():  # the same subject is in both datasets
+                            if not isinstance(vid1['os'][key], list):
+                                new_os = [vid1['os'][key]]
+                            else:
+                                new_os = vid1['os'][key]
+
+                            if isinstance(vid2['os'][key], list):
+                                new_dis_video['os'][key] = new_os + vid2['os'][key]
+                            else:
+                                new_os.append(vid2['os'][key])
+                                new_dis_video['os'][key] = new_os
+
+                    newone.dis_videos.append(new_dis_video)
+        return newone
+
+    def to_combined_overlap_dataset_file(self, output_dataset_filepath, second_dataset_reader, **kwargs):
+        combined_overlap_dataset = self.to_combined_overlap_dataset(second_dataset_reader, **kwargs)
+        self.write_out_dataset(combined_overlap_dataset, output_dataset_filepath)
+
+    def to_dictionary_style_dataset(self, **kwargs):
+        if isinstance(self.dis_videos[0]['os'], dict):
+            newone = self.dataset
+        else:
+
+            newone = self._prepare_new_dataset(kwargs)
+
+            # ref_videos: deepcopy
+            newone.ref_videos = copy.deepcopy(self.dataset.ref_videos)
+
+            # dis_videos: create a generic subject name for each entry in the list
+            newone.dis_videos = []
+            for dis_video in self.dis_videos:
+                new_dis_video = copy.deepcopy(dis_video)
+                new_dis_video['os'] = {}
+
+                for idx, score in enumerate(dis_video['os']):
+                    subj_name = self.dataset.dataset_name + '_subject' + str(idx)
+                    new_dis_video['os'][subj_name] = score
+
+                newone.dis_videos.append(new_dis_video)
+
+        return newone
+
+    def to_dictionary_style_dataset_file(self, dataset_filepath, **kwargs):
+        dict_style_dataset = self.to_dictionary_style_dataset()
+        self.write_out_dataset(dict_style_dataset, dataset_filepath)
+
 
 class MockedRawDatasetReader(RawDatasetReader):
 
     def __init__(self, dataset, **kwargs):
-        super(MockedRawDatasetReader, self).__init__(dataset)
         if 'input_dict' in kwargs:
             self.input_dict = kwargs['input_dict']
         else:
             self.input_dict = {}
+        super().__init__(dataset)
         self._assert_input_dict()
 
     def to_dataset(self):
@@ -480,11 +629,11 @@ class MockedRawDatasetReader(RawDatasetReader):
 
         # deep copy ref_videos and dis_videos
         newone.ref_videos = copy.deepcopy(self.dataset.ref_videos)
-        newone.dis_videos = copy.deepcopy(self.dataset.dis_videos)
+        newone.dis_videos = copy.deepcopy(self.dis_videos)
 
         # overwrite dis_video['os']
-        score_mtx = self.opinion_score_2darray
-        num_videos, num_subjects = score_mtx.shape
+        score_mtx = self.opinion_score_3darray
+        num_videos, num_subjects, max_repetitions = score_mtx.shape
         assert num_videos == len(newone.dis_videos)
         for scores, dis_video in zip(score_mtx, newone.dis_videos):
             dis_video['os'] = list(scores)
@@ -495,7 +644,7 @@ class MockedRawDatasetReader(RawDatasetReader):
 class SyntheticRawDatasetReader(MockedRawDatasetReader):
     """
     Dataset reader that generates synthetic data. It reads a dataset as baseline,
-    and override the opinion_score_2darray based on input_dict.
+    and override the opinion_score_3darray based on input_dict.
     """
 
     @property
@@ -526,9 +675,9 @@ class SyntheticRawDatasetReader(MockedRawDatasetReader):
             assert self.input_dict['seed'] is None or isinstance(self.input_dict['seed'], int)
 
     @property
-    def opinion_score_2darray(self):
+    def opinion_score_3darray(self):
         """
-        Override DatasetReader.opinion_score_2darray(self), based on input
+        Override DatasetReader.opinion_score_3darray(self), based on input
         synthetic_result.
         It follows the generative model:
         Z_e,s = Q_e + X_s + Y_[c(e)]
@@ -564,15 +713,18 @@ class SyntheticRawDatasetReader(MockedRawDatasetReader):
             assert len(phi_e) == E
             z_es += np.random.normal(0, 1, [E, S]) * np.tile(phi_e, (S, 1)).T
 
-        return z_es
+        z_es_3d = np.zeros([E, S, 1])
+        z_es_3d[:, :, 0] = z_es
+
+        return z_es_3d
 
 
 class SyntheticLogisticRawDatasetReader(SyntheticRawDatasetReader):
 
     @property
-    def opinion_score_2darray(self):
+    def opinion_score_3darray(self):
         """
-        Override DatasetReader.opinion_score_2darray(self), based on input
+        Override DatasetReader.opinion_score_3darray(self), based on input
         synthetic_result.
         use logistic instead of
         """
@@ -597,20 +749,23 @@ class SyntheticLogisticRawDatasetReader(SyntheticRawDatasetReader):
 
         assert 'quality_ambiguity' not in self.input_dict
 
-        return z_es
+        z_es_3d = np.zeros([E, S, 1])
+        z_es_3d[:, :, 0] = z_es
+
+        return z_es_3d
 
 
 class MissingDataRawDatasetReader(MockedRawDatasetReader):
     """
     Dataset reader that simulates random missing data. It reads a dataset as
-    baseline, and override the opinion_score_2darray based on input_dict.
+    baseline, and override the opinion_score_3darray based on input_dict.
     """
     def _assert_input_dict(self):
         assert 'missing_probability' in self.input_dict
 
     @property
-    def opinion_score_2darray(self):
-        score_mtx = super(MissingDataRawDatasetReader, self).opinion_score_2darray
+    def opinion_score_3darray(self):
+        score_mtx = super(MissingDataRawDatasetReader, self).opinion_score_3darray
 
         if 'seed' in self.input_dict:
             np.random.seed(self.input_dict['seed'])
@@ -624,7 +779,7 @@ class MissingDataRawDatasetReader(MockedRawDatasetReader):
 class SelectSubjectRawDatasetReader(MockedRawDatasetReader):
     """
     Dataset reader that only output selected subjects. It reads a dataset as a
-    baseline, and override the opinion_score_2darray and other fields based on
+    baseline, and override the opinion_score_3darray and other fields based on
     input_dict.
     """
     def _assert_input_dict(self):
@@ -642,28 +797,14 @@ class SelectSubjectRawDatasetReader(MockedRawDatasetReader):
         return len(self.input_dict['selected_subjects'])
 
     @property
-    def opinion_score_2darray(self):
+    def opinion_score_3darray(self):
         """
-        2darray storing raw opinion scores, with first dimension the number of
-        distorted videos, second dimension the number of observers
+        3darray storing raw opinion scores, with first dimension the number of
+        distorted videos, second dimension the number of observers, and third the maximum number of repetitions
         """
         selected_subjects = self.input_dict['selected_subjects']
-        score_mtx = np.zeros([self.num_dis_videos, self.num_observers])
-        for i_dis_video, dis_video in enumerate(self.dataset.dis_videos):
-
-            if isinstance(dis_video['os'], list):
-                score_mtx[i_dis_video, :] = np.array(dis_video['os'])[selected_subjects]
-
-            elif isinstance(dis_video['os'], dict):
-                selected_keys = np.array(list(dis_video['os'].keys()))[selected_subjects]
-                scores = []
-                for subject_key in selected_keys:
-                    scores.append(dis_video['os'][subject_key])
-                score_mtx[i_dis_video, :] = np.array(scores)
-
-            else:
-                raise AssertionError("The os in the dataset must be a dictionary or a list")
-
+        score_mtx = super().opinion_score_3darray
+        score_mtx = score_mtx[:, selected_subjects, :]
         return score_mtx
 
 
@@ -674,21 +815,17 @@ class SelectDisVideoRawDatasetReader(MockedRawDatasetReader):
 
         selected_dis_videos = self.input_dict['selected_dis_videos']
 
-        dis_video_idxs = range(super(SelectDisVideoRawDatasetReader, self).num_dis_videos)
+        dis_video_idxs = range(len(super().dis_videos))
         for dis_video in selected_dis_videos:
             assert dis_video in dis_video_idxs
 
     @property
-    def num_dis_videos(self):
-        return len(self.input_dict['selected_dis_videos'])
-
-    @property
-    def opinion_score_2darray(self):
-        selected_dis_videos = self.input_dict['selected_dis_videos']
-        score_mtx = np.zeros([self.num_dis_videos, self.num_observers])
-        for i_dis_video, dis_video in enumerate(selected_dis_videos):
-            score_mtx[i_dis_video, :] = np.array(self.dataset.dis_videos[dis_video]['os'])
-        return score_mtx
+    @override(DatasetReader)
+    def dis_videos(self):
+        d_assetid_disvideo = dict()  # build dict: asset_id -> dis_video
+        for dv in self.dataset.dis_videos:
+            d_assetid_disvideo[dv['asset_id']] = dv
+        return [d_assetid_disvideo[asset_id] for asset_id in self.input_dict['selected_dis_videos']]
 
     def to_dataset(self):
         raise NotImplementedError
@@ -697,7 +834,7 @@ class SelectDisVideoRawDatasetReader(MockedRawDatasetReader):
 class CorruptSubjectRawDatasetReader(MockedRawDatasetReader):
     """
     Dataset reader that have scores of selected subjects shuffled. It reads a
-    dataset as a baseline, and override the opinion_score_2darray and other
+    dataset as a baseline, and override the opinion_score_3darray and other
     fields based on input_dict.
     """
 
@@ -719,13 +856,13 @@ class CorruptSubjectRawDatasetReader(MockedRawDatasetReader):
             assert subject in observer_idxs
 
     @property
-    def opinion_score_2darray(self):
+    def opinion_score_3darray(self):
         """
-        2darray storing raw opinion scores, with first dimension the number of
-        distorted videos, second dimension the number of observers
+        3darray storing raw opinion scores, with first dimension the number of
+        distorted videos, second dimension the number of observers, and third the maximum of repetitions
         """
-        score_mtx = super(CorruptSubjectRawDatasetReader, self).opinion_score_2darray
-        num_video, num_subject = score_mtx.shape
+        score_mtx = super(CorruptSubjectRawDatasetReader, self).opinion_score_3darray
+        num_video, num_subject, max_repetitions = score_mtx.shape
 
         # for selected subjects, shuffle its score
         selected_subjects = self._get_selected_subjects()
@@ -738,9 +875,9 @@ class CorruptSubjectRawDatasetReader(MockedRawDatasetReader):
 
                 if corrupt_probability is not None:
                     videos = list(np.where(np.random.uniform(size=num_video) < corrupt_probability)[0])
-                    score_mtx[videos, subject] = np.random.permutation(score_mtx[videos, subject])
+                    score_mtx[videos, subject, :] = np.random.permutation(score_mtx[videos, subject, :])
                 else:
-                    np.random.shuffle(score_mtx[:, subject])
+                    np.random.shuffle(score_mtx[:, subject, :])
 
         elif corrupt_behavior == 'flip':
 
@@ -751,9 +888,9 @@ class CorruptSubjectRawDatasetReader(MockedRawDatasetReader):
 
                 if corrupt_probability is not None:
                     videos = list(np.where(np.random.uniform(size=num_video) < corrupt_probability)[0])
-                    score_mtx[videos, subject] = max_score + min_score - score_mtx[videos, subject]
+                    score_mtx[videos, subject, :] = max_score + min_score - score_mtx[videos, subject, :]
                 else:
-                    score_mtx[:, subject] = max_score + min_score - score_mtx[:, subject]
+                    score_mtx[:, subject, :] = max_score + min_score - score_mtx[:, subject, :]
 
         elif corrupt_behavior == 'min':
 
@@ -762,9 +899,9 @@ class CorruptSubjectRawDatasetReader(MockedRawDatasetReader):
 
                 if corrupt_probability is not None:
                     videos = list(np.where(np.random.uniform(size=num_video) < corrupt_probability)[0])
-                    score_mtx[videos, subject] = min_score
+                    score_mtx[videos, subject, :] = min_score
                 else:
-                    score_mtx[:, subject] = min_score
+                    score_mtx[:, subject, :] = min_score
 
         elif corrupt_behavior == 'mid':
 
@@ -776,9 +913,9 @@ class CorruptSubjectRawDatasetReader(MockedRawDatasetReader):
 
                 if corrupt_probability is not None:
                     videos = list(np.where(np.random.uniform(size=num_video) < corrupt_probability)[0])
-                    score_mtx[videos, subject] = mid_score
+                    score_mtx[videos, subject, :] = mid_score
                 else:
-                    score_mtx[:, subject] = mid_score
+                    score_mtx[:, subject, :] = mid_score
 
         elif corrupt_behavior == 'max':
 
@@ -787,9 +924,9 @@ class CorruptSubjectRawDatasetReader(MockedRawDatasetReader):
 
                 if corrupt_probability is not None:
                     videos = list(np.where(np.random.uniform(size=num_video) < corrupt_probability)[0])
-                    score_mtx[videos, subject] = max_score
+                    score_mtx[videos, subject, :] = max_score
                 else:
-                    score_mtx[:, subject] = max_score
+                    score_mtx[:, subject, :] = max_score
 
         elif corrupt_behavior == 'constant':
 
@@ -802,9 +939,9 @@ class CorruptSubjectRawDatasetReader(MockedRawDatasetReader):
 
                 if corrupt_probability is not None:
                     videos = list(np.where(np.random.uniform(size=num_video) < corrupt_probability)[0])
-                    score_mtx[videos, subject] = const_score
+                    score_mtx[videos, subject, :] = const_score
                 else:
-                    score_mtx[:, subject] = const_score
+                    score_mtx[:, subject, :] = const_score
 
         else:
             assert False
@@ -830,14 +967,14 @@ class CorruptDataRawDatasetReader(MockedRawDatasetReader):
 
     """
     Dataset reader that simulates random corrupted data. It reads a dataset as
-    baseline, and override the opinion_score_2darray based on input_dict.
+    baseline, and override the opinion_score_3darray based on input_dict.
     """
     def _assert_input_dict(self):
         assert 'corrupt_probability' in self.input_dict
 
     @property
-    def opinion_score_2darray(self):
-        score_mtx = super(CorruptDataRawDatasetReader, self).opinion_score_2darray
+    def opinion_score_3darray(self):
+        score_mtx = super(CorruptDataRawDatasetReader, self).opinion_score_3darray
 
         mask = np.random.uniform(size=score_mtx.shape)
         mask[mask > self.input_dict['corrupt_probability']] = 1.0
@@ -859,7 +996,7 @@ class PairedCompDatasetReader(RawDatasetReader):
         super(PairedCompDatasetReader, self)._assert_dataset()
 
         num_dis_videos = self.num_dis_videos
-        for dis_video in self.dataset.dis_videos:
+        for dis_video in self.dis_videos:
             # e.g. 'os': {(' Diana Pena Alas', 120): 1, ...
             assert 'os' in dis_video
             assert isinstance(dis_video['os'], dict)
@@ -886,20 +1023,23 @@ class PairedCompDatasetReader(RawDatasetReader):
 
         score_3darray = float("NaN") * np.ones([self.num_dis_videos, self.num_dis_videos, self.num_observers])
 
-        for i_dis_video, dis_video in enumerate(self.dataset.dis_videos):
+        for i_dis_video, dis_video in enumerate(self.dis_videos):
             for key, value in dis_video['os'].items():
+                assert value in [0, 0.5, 1], \
+                    f"expect value in [0, 0.5, 1], but got: {value}"
                 subject, pvs_j = key
                 pvs_i = i_dis_video
-                score_3darray[pvs_i][pvs_j][dict_observer_to_iobserver[subject]] = value # CAUTION: note the dimension change!
+                score_3darray[pvs_i][pvs_j][dict_observer_to_iobserver[subject]] = value  # CAUTION: note the dimension change!
+                score_3darray[pvs_j][pvs_i][dict_observer_to_iobserver[subject]] = 1.0 - value
 
         return score_3darray
 
     def _get_list_observers(self):
-        for dis_video in self.dataset.dis_videos:
+        for dis_video in self.dis_videos:
             assert isinstance(dis_video['os'], dict)
 
         list_observers = []
-        for dis_video in self.dataset.dis_videos:
+        for dis_video in self.dis_videos:
             observers = map(lambda x: x[0], dis_video['os'].keys())
             observers = list(set(observers))
             list_observers += observers
@@ -910,10 +1050,8 @@ class PairedCompDatasetReader(RawDatasetReader):
     def ref_score(self):
         raise NotImplementedError
 
-    @property
-    def opinion_score_2darray(self):
-        # return np.nansum(self.opinion_score_3darray, axis=1)
-        raise NotImplementedError
-
     def to_persubject_dataset(self, quality_scores, **kwargs):
         raise NotImplementedError
+
+    def to_combined_overlap_dataset(self, second_dataset_filepath, **kwargs):
+        raise NotImplementedError
